/*
-----------------------------------------------------------------------------
 File: RageScreen.h

 Desc: Light wrapper around D3D and utility functions for drawing quads.

 Copyright (c) 2001 Chris Danford.  All rights reserved.
-----------------------------------------------------------------------------
*/

class RageScreen;
typedef RageScreen* LPRageScreen;



#ifndef _RageSCREEN_H_
#define _RageSCREEN_H_

#include "RageTexture.h"
#include <D3DX8.h>

// A structure for our custom vertex type. We added texture coordinates
struct CUSTOMVERTEX
{
    D3DXVECTOR3 p;			// The position
    D3DCOLOR    color;		// The color
    FLOAT       tu, tv;		// The texture coordinates
};

// Our custom FVF, which describes our custom vertex structure
#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZ|D3DFVF_DIFFUSE|D3DFVF_TEX1)


const int MAX_NUM_VERTICIES = 1000*4;
const int MAX_NUM_QUADS = MAX_NUM_VERTICIES/4;


class RageScreen
{
public:
	RageScreen( HWND hWnd );
	~RageScreen();
	BOOL SwitchDisplayMode( BOOL bWindowed, DWORD dwWidth, DWORD dwHeight, DWORD dwBPP );

	LPDIRECT3D8 GetD3D();
	LPDIRECT3DDEVICE8 GetDevice();

	HRESULT Reset();
	HRESULT BeginFrame();
	HRESULT EndFrame();
	HRESULT ShowFrame();

	HRESULT Invalidate();
	HRESULT Restore();


	BOOL IsWindowed()	{ return m_d3dpp.Windowed; };
	DWORD GetWidth()	{ return m_d3dpp.BackBufferWidth; };
	DWORD GetHeight()	{ return m_d3dpp.BackBufferHeight; };
	DWORD GetBPP()		
	{ 
		switch( m_d3dpp.BackBufferFormat )
		{
		case D3DFMT_R5G6B5:
		case D3DFMT_X1R5G5B5:
		case D3DFMT_A1R5G5B5:
			return 16;
		case D3DFMT_R8G8B8:
		case D3DFMT_X8R8G8B8:
		case D3DFMT_A8R8G8B8:
			return 32;
		default:
			ASSERT( false );	// unexpected format
			return 0;
		}
	}

	


	// functions for drawing rectangles
	void CreateVertexBuffer();
	void ReleaseVertexBuffer();

	LPDIRECT3DVERTEXBUFFER8 GetVertexBuffer() { return m_pVB; };

	void ResetMatrixStack() 
	{ 
		m_MatrixStack.SetSize( 1, 20 );
		D3DXMatrixIdentity( &GetTopMatrix() );
		
		m_pd3dDevice->SetTransform( D3DTS_WORLD, &m_MatrixStack[m_MatrixStack.GetSize()-1] );
	};
	void PushMatrix() 
	{ 
		m_MatrixStack.Add( GetTopMatrix() );	
		ASSERT(m_MatrixStack.GetSize()<30);		// check for infinite loop
		m_pd3dDevice->SetTransform( D3DTS_WORLD, &m_MatrixStack[m_MatrixStack.GetSize()-1] );
	};
	void PopMatrix() 
	{ 
		m_MatrixStack.RemoveAt( m_MatrixStack.GetSize()-1 ); 
		m_pd3dDevice->SetTransform( D3DTS_WORLD, &m_MatrixStack[m_MatrixStack.GetSize()-1] );
	};
	void Translate( FLOAT x, FLOAT y, FLOAT z )
	{
		D3DXMATRIX matTemp;
		D3DXMatrixTranslation( &matTemp, x, y, z );
		D3DXMATRIX& matTop = GetTopMatrix();
		matTop = matTemp * matTop;
		m_pd3dDevice->SetTransform( D3DTS_WORLD, &matTop ); 
	};
	void Scale( FLOAT x, FLOAT y, FLOAT z )
	{
		D3DXMATRIX matTemp;
		D3DXMatrixScaling( &matTemp, x, y, z );
		D3DXMATRIX& matTop = GetTopMatrix();
		matTop = matTemp * matTop;
		m_pd3dDevice->SetTransform( D3DTS_WORLD, &matTop ); 
	};
	void RotateZ( FLOAT z )
	{
		D3DXMATRIX matTemp;
		D3DXMatrixRotationZ( &matTemp, z );
		D3DXMATRIX& matTop = GetTopMatrix();
		matTop = matTemp * matTop;
		m_pd3dDevice->SetTransform( D3DTS_WORLD, &matTop ); 
	};
	void RotateYawPitchRoll( FLOAT x, FLOAT y, FLOAT z )
	{
		D3DXMATRIX matTemp;
		D3DXMatrixRotationYawPitchRoll( &matTemp, x, y, z );
		D3DXMATRIX& matTop = GetTopMatrix();
		matTop = matTemp * matTop;
		m_pd3dDevice->SetTransform( D3DTS_WORLD, &matTop ); 
	};

	float GetFPS() { return m_fFPS; };

private:
	D3DXMATRIX& GetTopMatrix() { return m_MatrixStack.ElementAt( m_MatrixStack.GetSize()-1 ); };

	HWND m_hWnd;

	// DirectDraw/Direct3D objects
	LPDIRECT3D8			m_pd3d;
	LPDIRECT3DDEVICE8	m_pd3dDevice;

	D3DDISPLAYMODE		m_DesktopMode;
	D3DPRESENT_PARAMETERS	m_d3dpp;

	// a vertex buffer for all to share
	LPDIRECT3DVERTEXBUFFER8 m_pVB;

	// OpenGL-like matrix stack
	CArray<D3DXMATRIX, D3DXMATRIX&>	m_MatrixStack;

	// for performance stats
	DWORD m_dwLastUpdateTicks;
	int m_iFramesRenderedSinceLastCheck;
	float m_fFPS;
};


extern LPRageScreen			SCREEN;	// global and accessable from anywhere in our program


#endif
