#include "stdafx.h"
/*
-----------------------------------------------------------------------------
 File: Steps.h

 Desc: Hold steps data and metadata.  Does not hold dancing-time information,
	   like keeping track of which steps have been stepped on (PlayerSteps does that).

 Copyright (c) 2001 Chris Danford.  All rights reserved.
-----------------------------------------------------------------------------
*/

#include "Song.h"
#include "Steps.h"
#include "Util.h"
#include "IniFile.h"
#include "math.h"	// for fabs()
#include "RageUtil.h"

#include "Steps.h"



Steps::Steps()
{
	m_GameMode = MODE_SINGLE;
	m_DifficultyClass = CLASS_EASY;
	m_iNumFeet = 0;
	for( int i=0; i<MAX_TAP_STEP_ELEMENTS; i++ )
		m_TapSteps[i] = STEP_NONE;

	m_iNumTimesPlayed = 0;
	m_iMaxCombo = 0;
	m_iTopScore = 0;
	m_TopGrade = GRADE_NO_DATA;
}


bool Steps::LoadStepsFromBMSFile( CString sPath )
{
	m_sStepsFilePath = sPath;

// BMS encoding:
// 4&8panel: Player1  Player2
//    Left		11		21
//    Down		13		23
//    Up		15		25
//    Right		16		26
//	6panel:	 Player1  Player2
//    Left		11		21
//    Left+Up	12		22
//    Down		13		23
//    Up		14		24
//    Up+Right	15		25
//    Right		16		26
//
//	Notice that 15 and 25 have two different meanings!  What were they thinking???
//	While reading in, use the 6 panel mapping.  After reading in, detect if only 4 notes
//	are filled in.  If so, shift the Up+Right column back to the Up column
//
// Our internal encoding:
// 0x0001 left		player 1
// 0x0002 left+up	player 1
// 0x0004 down		player 1
// 0x0008 up		player 1
// 0x0010 up+right	player 1
// 0x0020 right		player 1
// 0x0040 left		player 2
// 0x0080 left+up	player 2
// 0x0100 down		player 2
// 0x0200 up		player 2
// 0x0400 up+right	player 2
// 0x0800 right		player 2
CMap<int, int, TapStep, TapStep>  mapBMSTrackNumToOurStepBit;
mapBMSTrackNumToOurStepBit[11] = STEP_PAD1_LEFT;
mapBMSTrackNumToOurStepBit[12] = STEP_PAD1_UPLEFT;
mapBMSTrackNumToOurStepBit[13] = STEP_PAD1_DOWN;
mapBMSTrackNumToOurStepBit[14] = STEP_PAD1_UP;
mapBMSTrackNumToOurStepBit[15] = STEP_PAD1_UPRIGHT;
mapBMSTrackNumToOurStepBit[16] = STEP_PAD1_RIGHT;
mapBMSTrackNumToOurStepBit[21] = STEP_PAD2_LEFT;
mapBMSTrackNumToOurStepBit[22] = STEP_PAD2_UPLEFT;
mapBMSTrackNumToOurStepBit[23] = STEP_PAD2_DOWN;
mapBMSTrackNumToOurStepBit[24] = STEP_PAD2_UP;
mapBMSTrackNumToOurStepBit[25] = STEP_PAD2_UPRIGHT;
mapBMSTrackNumToOurStepBit[26] = STEP_PAD2_RIGHT;



	CStdioFile file;	
	if( !file.Open( sPath, CFile::modeRead|CFile::shareDenyNone ) )
	{
		RageError( ssprintf("Failed to open %s.", sPath) );
		return false;
	}

	CString line;
	while( file.ReadString(line) )	// foreach line
	{
		CString value_name;		// fill these in
		CString value_data;	

		// BMS value names can be separated by a space or a colon.
		int iIndexOfFirstColon = line.Find( ":" );
		int iIndexOfFirstSpace = line.Find( " " );

		if( iIndexOfFirstColon == -1 )
			iIndexOfFirstColon = 10000;
		if( iIndexOfFirstSpace == -1 )
			iIndexOfFirstSpace = 10000;
		
		int iIndexOfSeparator = min( iIndexOfFirstSpace, iIndexOfFirstColon );

		if( iIndexOfSeparator != 10000 )
		{
			value_name = line.Mid( 0, iIndexOfSeparator );
			value_data = line;	// the rest
			value_data.Delete(0,iIndexOfSeparator+1);
		}
		else	// no separator
		{
			value_name = line;
		}

		value_name.MakeLower();

		if( -1 != value_name.Find("#player") ) 
		{
			switch( atoi(value_data) )
			{
			case 1:		// 4 or 6 single
				m_GameMode = MODE_SINGLE;
				// if the mode should be solo, then we'll update m_GameMode below when we read in step data
				break;
			case 2:		// couple/battle
				m_GameMode = MODE_COUPLE;
				break;
			case 3:		// double
				m_GameMode = MODE_DOUBLE;
				break;
			}
		}
		if( -1 != value_name.Find("#title") )
		{
			m_sDescription = value_data;
			
			// extract the steps description (looks like 'Music <BASIC>')
			int iPosOpenBracket = m_sDescription.Find( "<" );
			if( iPosOpenBracket == -1 )
				iPosOpenBracket = m_sDescription.Find( "(" );
			int iPosCloseBracket = m_sDescription.Find( ">" );
			if( iPosCloseBracket == -1 )
				iPosCloseBracket = m_sDescription.Find( ")" );

			if( iPosOpenBracket != -1  &&  iPosCloseBracket != -1 )
				m_sDescription = m_sDescription.Mid( iPosOpenBracket+1, iPosCloseBracket-iPosOpenBracket-1 );
			m_sDescription.MakeLower();
			RageLog( "Steps description found to be '%s'", m_sDescription );

			// if there's a 6 in the description, it's probably part of "6panel" or "6-panel"
			if( m_sDescription.Find("6") != -1 )
				m_GameMode = MODE_SOLO;
			
		}
		if( -1 != value_name.Find("#playlevel") ) 
		{
			m_iNumFeet = atoi( value_data );
		}
		else if( value_name.Left(1) == "#"  
			 && IsAnInt( value_name.Mid(1,3) )
			 && IsAnInt( value_name.Mid(4,2) ) )	// this is step or offset data.  Looks like "#00705"
		{
			int iMeasureNo	= atoi( value_name.Mid(1,3) );
			int iTrackNum	= atoi( value_name.Mid(4,2) );

			CString sNoteData = value_data;
			CArray<bool, bool&> arrayNotes;

			for( int i=0; i<sNoteData.GetLength(); i+=2 )
			{
				bool bThisIsANote = sNoteData.Mid(i,2) != "00";
				arrayNotes.Add( bThisIsANote );
			}

			const int iNumNotesInThisMeasure = arrayNotes.GetSize();
			//RageLog( "%s:%s: iMeasureNo = %d, iNoteNum = %d, iNumNotesInThisMeasure = %d", 
			//	valuename, sNoteData, iMeasureNo, iNoteNum, iNumNotesInThisMeasure );
			for( int j=0; j<iNumNotesInThisMeasure; j++ )
			{
				if( arrayNotes.GetAt(j) == TRUE )
				{
					float fPercentThroughMeasure = (float)j/(float)iNumNotesInThisMeasure;

					// index is in quarter beats starting at beat 0
					int iStepIndex = (int) ( (iMeasureNo + fPercentThroughMeasure)
									 * BEATS_PER_MEASURE * ELEMENTS_PER_BEAT );

					m_TapSteps[iStepIndex] |= mapBMSTrackNumToOurStepBit[iTrackNum];

				}
			}
		}
	}
	
	switch( m_GameMode  )	// if there are 4 panels, then we have to flip the Up and Up+Right bits
	{
	case MODE_SINGLE:
	case MODE_DOUBLE:
	case MODE_COUPLE:
		for( int i=0; i<MAX_TAP_STEP_ELEMENTS; i++ )
		{
			if( m_TapSteps[i] & STEP_PAD1_UPRIGHT )	// if up+right is a step
			{
				m_TapSteps[i] |= STEP_PAD1_UP;			// add up
				m_TapSteps[i] &= ~STEP_PAD1_UPRIGHT;	// subtract up+right
			}
			if( m_TapSteps[i] & STEP_PAD2_UPRIGHT )	// if up+right is a step
			{
				m_TapSteps[i] |= STEP_PAD2_UP;			// add up
				m_TapSteps[i] &= ~STEP_PAD2_UPRIGHT;	// subtract up+right
			}
		}
		break;
	}

	// we're done reading in all of the BMS values

	CArray<TapStep, TapStep>	arrayStepTypes;
	arrayStepTypes.Add( STEP_PAD1_LEFT );
	arrayStepTypes.Add( STEP_PAD1_UPLEFT );
	arrayStepTypes.Add( STEP_PAD1_DOWN );
	arrayStepTypes.Add( STEP_PAD1_UP );
	arrayStepTypes.Add( STEP_PAD1_UPRIGHT );
	arrayStepTypes.Add( STEP_PAD1_RIGHT );
	arrayStepTypes.Add( STEP_PAD2_LEFT );
	arrayStepTypes.Add( STEP_PAD2_UPLEFT );
	arrayStepTypes.Add( STEP_PAD2_DOWN );
	arrayStepTypes.Add( STEP_PAD2_UP );
	arrayStepTypes.Add( STEP_PAD2_UPRIGHT );
	arrayStepTypes.Add( STEP_PAD2_RIGHT );



	// search for runs of 32nd notes of all the same note type, and convert them to freezes
	for( int j=0; j<arrayStepTypes.GetSize(); j++ )		// for each type of step
	{
		TapStep &stepType = arrayStepTypes[j];

		int iIndexRunStart = -1;	// -1 means we're not in the middle of a run

		for( float fBeat=0; fBeat<MAX_BEATS; fBeat += 1/32.0f )		// foreach 32nd note
		{
			int i = BeatToStepIndex( fBeat );

			if( m_TapSteps[i] & stepType )	// there is a step here
			{
				if( iIndexRunStart == -1 )	// we are not in the middle of a run
				{
					iIndexRunStart = i;
				}
				else	// we are in the middle of a run
				{
					;	// do nothing.  Keep reading in until we hit the end of the run.	
				}
			}
			else	// there isn't a step here
			{
				if( iIndexRunStart == -1 )	// we are not in the middle of a run
				{
					;	// do nothing
				}
				else	// we are in the middle of a run
				{
					// this ends the run

 					if( i - iIndexRunStart < 2 )	// reject runs of length 1
					{
						;	// do nothing
					}
					else	// this run is longer than 1
					{
						// add this HoldStep to the list
						int iIndexRunEnd = i - 1;
						HoldStep hs = { stepType, iIndexRunStart, iIndexRunEnd };
						m_HoldSteps[m_iNumHoldSteps++] = hs;

						// clear out all the 32nd notes in this run from m_TapSteps
						for( int s=iIndexRunStart; s<=iIndexRunEnd; s++ )
							m_TapSteps[s] &= ~stepType;	// subtract the current step type


						RageLog( "Found a run with Step %d: start: %d, end: %d.", hs.m_TapStep, hs.m_iStartIndex, hs.m_iEndIndex );
					}

					iIndexRunStart = -1;	// reset so we can look for more runs
				}
			}
		}
		// done looking for runs


	}


	return true;
}


TapStep DWIcharToStepLeftSide( char c )
{
	switch( c )
		{
		case '0':	return STEP_NONE;
		case '1':	return STEP_PAD1_DOWN | STEP_PAD1_LEFT;
		case '2':	return STEP_PAD1_DOWN;
		case '3':	return STEP_PAD1_DOWN | STEP_PAD1_RIGHT;
		case '4':	return STEP_PAD1_LEFT;
		case '5':	return STEP_NONE;
		case '6':	return STEP_PAD1_RIGHT;
		case '7':	return STEP_PAD1_UP | STEP_PAD1_LEFT;
		case '8':	return STEP_PAD1_UP;
		case '9':	return STEP_PAD1_UP | STEP_PAD1_RIGHT;
		case 'A':	return STEP_PAD1_UP | STEP_PAD1_DOWN;
		case 'B':	return STEP_PAD1_LEFT | STEP_PAD1_RIGHT;
		default:	RageError( ssprintf("Encountered invalid DWI note characer '%c'", c) );
			return STEP_NONE;
	}

}

TapStep DWIcharToStepRightSide( char c )
{
	switch( c )
		{
		case '0':	return STEP_NONE;
		case '1':	return STEP_PAD2_DOWN | STEP_PAD2_LEFT;
		case '2':	return STEP_PAD2_DOWN;
		case '3':	return STEP_PAD2_DOWN | STEP_PAD2_RIGHT;
		case '4':	return STEP_PAD2_LEFT;
		case '5':	return STEP_NONE;
		case '6':	return STEP_PAD2_RIGHT;
		case '7':	return STEP_PAD2_UP | STEP_PAD2_LEFT;
		case '8':	return STEP_PAD2_UP;
		case '9':	return STEP_PAD2_UP | STEP_PAD2_RIGHT;
		case 'A':	return STEP_PAD2_UP | STEP_PAD2_DOWN;
		case 'B':	return STEP_PAD2_LEFT | STEP_PAD2_RIGHT;
		default:	RageError( ssprintf("Encountered invalid DWI note characer '%c'", c) );
			return STEP_NONE;
	}

}

bool Steps::LoadFromDWIValueTokens( CStringArray &arrayValueTokens )
{
	m_sStepsFilePath = "";	 // what could we put here?
	
	if(		 arrayValueTokens[0] == "#SINGLE" )
		m_GameMode = MODE_SINGLE;
	else if( arrayValueTokens[0] == "#DOUBLE" )
		m_GameMode = MODE_DOUBLE;
	else if( arrayValueTokens[0] == "#COUPLE" )
		m_GameMode = MODE_COUPLE;

	m_sDescription = arrayValueTokens[1];

	if( m_sDescription == "BASIC" )
		m_DifficultyClass = CLASS_EASY;
	if( m_sDescription == "ANOTHER" )
		m_DifficultyClass = CLASS_MEDIUM;
	if( m_sDescription == "MANIAC" )
		m_DifficultyClass = CLASS_HARD;
	else
		m_DifficultyClass = CLASS_OTHER;

	m_sMapper = "";

	m_iNumFeet = atoi( arrayValueTokens[2] );


	CString &sStepData = arrayValueTokens[3];
	double fCurrentBeat = 0;
	double fCurrentIncrementer = 1.0/8 * BEATS_PER_MEASURE;
	for( int i=0; i<sStepData.GetLength(); )
	{
		char c = sStepData[i++];

		switch( c )
		{
		// begins a series
		case '(':
			fCurrentIncrementer = 1.0/16 * BEATS_PER_MEASURE;
			break;
		case '[':
			fCurrentIncrementer = 1.0/24 * BEATS_PER_MEASURE;
			break;
		case '{':
			fCurrentIncrementer = 1.0/64 * BEATS_PER_MEASURE;
			break;
		case '<':
			fCurrentIncrementer = 1.0/192 * BEATS_PER_MEASURE;
			break;

		// ends a series
		case ')':
		case ']':
		case '}':
		case '>':
			fCurrentIncrementer = 1.0/8 * BEATS_PER_MEASURE;
			break;

		case ' ':
			break;	// do nothing!
		
		case '!':		// hold start
		{

			// rewind and get the last step we inserted
			double fLastStepBeat = fCurrentBeat - fCurrentIncrementer;
			int iLastStepIndex = BeatToStepIndex( (float)fLastStepBeat );
			TapStep &lastStep = m_TapSteps[iLastStepIndex];

			char holdChar = sStepData[i++];
			TapStep holdStep = DWIcharToStepLeftSide( holdChar );
			lastStep &= ~holdStep;		// subtract this Hold Step from the last Step

			int iStepIndex = BeatToStepIndex( (float)fLastStepBeat );

			HoldStep hs = { 0, iStepIndex, -1 };	// end index is -1.  We need to fill this in later!

			if( holdStep & STEP_PAD1_LEFT )
			{
				// check to see if there is another HoldStep that ends on this same beat
				HoldStep* pClosedHoldStepThatEndsAtTheSameTime = NULL;
				for( int h=0; h<m_iNumHoldSteps; h++ )
				{
					HoldStep &hs = m_HoldSteps[h];
					if( (STEP_PAD1_LEFT & hs.m_TapStep)  &&  hs.m_iEndIndex == iStepIndex )
						pClosedHoldStepThatEndsAtTheSameTime = &m_HoldSteps[h];
				}
				if( pClosedHoldStepThatEndsAtTheSameTime != NULL )
				{
					pClosedHoldStepThatEndsAtTheSameTime->m_iEndIndex = -1;	// open it back up
					//RageLog( "Re-opening the HoldStep at %d (%f)", iLastStepIndex, fLastStepBeat );
				}
				else
				{
					hs.m_TapStep = STEP_PAD1_LEFT;
					m_HoldSteps[m_iNumHoldSteps++] = hs;
					//RageLog( "Found a hold step at %d (%f).  The Step is now %d and the HoldStep is %d (%c)", iLastStepIndex, fLastStepBeat, lastStep, hs.m_TapStep, holdChar );
				}
			}
			if( holdStep & STEP_PAD1_RIGHT )
			{
				// check to see if there is another HoldStep that ends on this same beat
				HoldStep* pClosedHoldStepThatEndsAtTheSameTime = NULL;
				for( int h=0; h<m_iNumHoldSteps; h++ )
				{
					HoldStep &hs = m_HoldSteps[h];
					if( (STEP_PAD1_RIGHT & hs.m_TapStep)  &&  hs.m_iEndIndex == iStepIndex )
						pClosedHoldStepThatEndsAtTheSameTime = &m_HoldSteps[h];
				}
				if( pClosedHoldStepThatEndsAtTheSameTime != NULL )
				{
					pClosedHoldStepThatEndsAtTheSameTime->m_iEndIndex = -1;	// open it back up
					//RageLog( "Re-opening the HoldStep at %d (%f)", iLastStepIndex, fLastStepBeat );
				}
				else
				{
					hs.m_TapStep = STEP_PAD1_RIGHT;
					m_HoldSteps[m_iNumHoldSteps++] = hs;
					//RageLog( "Found a hold step at %d (%f).  The Step is now %d and the HoldStep is %d (%c)", iLastStepIndex, fLastStepBeat, lastStep, hs.m_TapStep, holdChar );
				}
			}
			if( holdStep & STEP_PAD1_UP )
			{
				// check to see if there is another HoldStep that ends on this same beat
				HoldStep* pClosedHoldStepThatEndsAtTheSameTime = NULL;
				for( int h=0; h<m_iNumHoldSteps; h++ )
				{
					HoldStep &hs = m_HoldSteps[h];
					if( (STEP_PAD1_UP & hs.m_TapStep)  &&  hs.m_iEndIndex == iStepIndex )
						pClosedHoldStepThatEndsAtTheSameTime = &m_HoldSteps[h];
				}
				if( pClosedHoldStepThatEndsAtTheSameTime != NULL )
				{
					pClosedHoldStepThatEndsAtTheSameTime->m_iEndIndex = -1;	// open it back up
					//RageLog( "Re-opening the HoldStep at %d (%f)", iLastStepIndex, fLastStepBeat );
				}
				else
				{
					hs.m_TapStep = STEP_PAD1_UP;
					m_HoldSteps[m_iNumHoldSteps++] = hs;
					//RageLog( "Found a hold step at %d (%f).  The Step is now %d and the HoldStep is %d (%c)", iLastStepIndex, fLastStepBeat, lastStep, hs.m_TapStep, holdChar );
				}
			}
			if( holdStep & STEP_PAD1_DOWN )
			{
				// check to see if there is another HoldStep that ends on this same beat
				HoldStep* pClosedHoldStepThatEndsAtTheSameTime = NULL;
				for( int h=0; h<m_iNumHoldSteps; h++ )
				{
					HoldStep &hs = m_HoldSteps[h];
					if( (STEP_PAD1_DOWN & hs.m_TapStep)  &&  hs.m_iEndIndex == iStepIndex )
						pClosedHoldStepThatEndsAtTheSameTime = &m_HoldSteps[h];
				}
				if( pClosedHoldStepThatEndsAtTheSameTime != NULL )
				{
					pClosedHoldStepThatEndsAtTheSameTime->m_iEndIndex = -1;	// open it back up
					//RageLog( "Re-opening the HoldStep at %d (%f)", iLastStepIndex, fLastStepBeat );
				}
				else
				{
					hs.m_TapStep = STEP_PAD1_DOWN;
					m_HoldSteps[m_iNumHoldSteps++] = hs;
					//RageLog( "Found a hold step at %d (%f).  The Step is now %d and the HoldStep is %d (%c)", iLastStepIndex, fLastStepBeat, lastStep, hs.m_TapStep, holdChar );
				}
			}
			break;
		}
/*		case '<':
			// hold segment
			// Handle this by just inserting one step at the end of the hold note.  
			// Then our logic below will find it an complete the hold step.

			char charStep;
			charStep = sStepData[i++];
			fCurrentBeat += 1.0/192 * BEATS_PER_MEASURE;

			while( c != '>' )	// keep reading in characters until we reach the end of the hold segment
			{
				c = sStepData[i++];
				fCurrentBeat += 1.0/192 * BEATS_PER_MEASURE;
			}

			c = charStep;	// put the step character back into 'c', and let the handler below do its magic
			// fall through!

			// hack!  undo the beat increment below
			fCurrentBeat -= fCurrentIncrementer;
*/
		default:	// this is a note character

			TapStep step = DWIcharToStepLeftSide( c );

			int iStepIndex = BeatToStepIndex( (float)fCurrentBeat );
			m_TapSteps[iStepIndex] |= step;

			//RageLog( "Setting m_TapSteps[%d] (%f) = %d  (DWI: %c)", iStepIndex, fCurrentBeat, step, c );

			// check to see if this ends one of the open HoldSteps
			for( int h=0; h<m_iNumHoldSteps; h++ )
			{
				HoldStep &hs = m_HoldSteps[h];
				if( hs.m_iEndIndex != -1 )	// this HoldStep is already finished
					continue;				// continue on to the next HoldStep

				if( step & hs.m_TapStep )	// if this step covers this hold step
				{
					hs.m_iEndIndex = iStepIndex;	// end the HoldStep here

					//RageLog( "HoldStep: step=%d, start=%d, end=%d", hs.m_TapStep, hs.m_iStartIndex, hs.m_iEndIndex );

					//RageLog( "m_TapSteps[%d]: subtracting %d from %d to make %d", iStepIndex, hs.m_TapStep, m_TapSteps[iStepIndex], m_TapSteps[iStepIndex] & ~hs.m_TapStep );
					m_TapSteps[iStepIndex] = m_TapSteps[iStepIndex] & ~hs.m_TapStep;	// subtract out the hold step
				}
			}
			
			fCurrentBeat += fCurrentIncrementer;
			break;
		}
		
	}

	if( m_GameMode == MODE_DOUBLE )
	{
		//RageLog( "Loading the right side Steps for this double pattern." );

		CString &sStepData = arrayValueTokens[4];
		double fCurrentBeat = 0;
		double fCurrentIncrementer = 1.0/8 * BEATS_PER_MEASURE;
		for( int i=0; i<sStepData.GetLength(); )
		{
			char c = sStepData[i++];

			switch( c )
			{
			// begins a series
			case '(':
				fCurrentIncrementer = 1.0/16 * BEATS_PER_MEASURE;
				break;
			case '[':
				fCurrentIncrementer = 1.0/24 * BEATS_PER_MEASURE;
				break;
			case '{':
				fCurrentIncrementer = 1.0/64 * BEATS_PER_MEASURE;
				break;

			// ends a series
			case ')':
			case ']':
			case '}':
				fCurrentIncrementer = 1.0/8 * BEATS_PER_MEASURE;
				break;

			case ' ':
				break;	// do nothing!
			
			case '!':		// hold start
			{

				// rewind and get the last step we inserted
				double fLastStepBeat = fCurrentBeat - fCurrentIncrementer;
				int iLastStepIndex = BeatToStepIndex( (float)fLastStepBeat );
				TapStep &lastStep = m_TapSteps[iLastStepIndex];

				char holdChar = sStepData[i++];
				TapStep holdStep = DWIcharToStepRightSide( holdChar );
				lastStep &= ~holdStep;		// subtract this Hold Step from the last Step

				int iStepIndex = BeatToStepIndex( (float)fLastStepBeat );

				HoldStep hs = { 0, iStepIndex, -1 };	// end index is -1.  We need to fill this in later!

				if( holdStep & STEP_PAD2_LEFT )
				{
					// check to see if there is another HoldStep that ends on this same beat
					HoldStep* pClosedHoldStepThatEndsAtTheSameTime = NULL;
					for( int h=0; h<m_iNumHoldSteps; h++ )
					{
						HoldStep &hs = m_HoldSteps[h];
						if( (STEP_PAD2_LEFT & hs.m_TapStep)  &&  hs.m_iEndIndex == iStepIndex )
							pClosedHoldStepThatEndsAtTheSameTime = &m_HoldSteps[h];
					}
					if( pClosedHoldStepThatEndsAtTheSameTime != NULL )
					{
						pClosedHoldStepThatEndsAtTheSameTime->m_iEndIndex = -1;	// open it back up
						//RageLog( "Re-opening the HoldStep at %d (%f)", iLastStepIndex, fLastStepBeat );
					}
					else
					{
						hs.m_TapStep = STEP_PAD2_LEFT;
						m_HoldSteps[m_iNumHoldSteps++] = hs;
						//RageLog( "Found a hold step at %d (%f).  The Step is now %d and the HoldStep is %d (%c)", iLastStepIndex, fLastStepBeat, lastStep, hs.m_TapStep, holdChar );
					}
				}
				if( holdStep & STEP_PAD2_RIGHT )
				{
					// check to see if there is another HoldStep that ends on this same beat
					HoldStep* pClosedHoldStepThatEndsAtTheSameTime = NULL;
					for( int h=0; h<m_iNumHoldSteps; h++ )
					{
						HoldStep &hs = m_HoldSteps[h];
						if( (STEP_PAD2_RIGHT & hs.m_TapStep)  &&  hs.m_iEndIndex == iStepIndex )
							pClosedHoldStepThatEndsAtTheSameTime = &m_HoldSteps[h];
					}
					if( pClosedHoldStepThatEndsAtTheSameTime != NULL )
					{
						pClosedHoldStepThatEndsAtTheSameTime->m_iEndIndex = -1;	// open it back up
						//RageLog( "Re-opening the HoldStep at %d (%f)", iLastStepIndex, fLastStepBeat );
					}
					else
					{
						hs.m_TapStep = STEP_PAD2_RIGHT;
						m_HoldSteps[m_iNumHoldSteps++] = hs;
						//RageLog( "Found a hold step at %d (%f).  The Step is now %d and the HoldStep is %d (%c)", iLastStepIndex, fLastStepBeat, lastStep, hs.m_TapStep, holdChar );
					}
				}
				if( holdStep & STEP_PAD2_UP )
				{
					// check to see if there is another HoldStep that ends on this same beat
					HoldStep* pClosedHoldStepThatEndsAtTheSameTime = NULL;
					for( int h=0; h<m_iNumHoldSteps; h++ )
					{
						HoldStep &hs = m_HoldSteps[h];
						if( (STEP_PAD2_UP & hs.m_TapStep)  &&  hs.m_iEndIndex == iStepIndex )
							pClosedHoldStepThatEndsAtTheSameTime = &m_HoldSteps[h];
					}
					if( pClosedHoldStepThatEndsAtTheSameTime != NULL )
					{
						pClosedHoldStepThatEndsAtTheSameTime->m_iEndIndex = -1;	// open it back up
						//RageLog( "Re-opening the HoldStep at %d (%f)", iLastStepIndex, fLastStepBeat );
					}
					else
					{
						hs.m_TapStep = STEP_PAD2_UP;
						m_HoldSteps[m_iNumHoldSteps++] = hs;
						//RageLog( "Found a hold step at %d (%f).  The Step is now %d and the HoldStep is %d (%c)", iLastStepIndex, fLastStepBeat, lastStep, hs.m_TapStep, holdChar );
					}
				}
				if( holdStep & STEP_PAD2_DOWN )
				{
					// check to see if there is another HoldStep that ends on this same beat
					HoldStep* pClosedHoldStepThatEndsAtTheSameTime = NULL;
					for( int h=0; h<m_iNumHoldSteps; h++ )
					{
						HoldStep &hs = m_HoldSteps[h];
						if( (STEP_PAD2_DOWN & hs.m_TapStep)  &&  hs.m_iEndIndex == iStepIndex )
							pClosedHoldStepThatEndsAtTheSameTime = &m_HoldSteps[h];
					}
					if( pClosedHoldStepThatEndsAtTheSameTime != NULL )
					{
						pClosedHoldStepThatEndsAtTheSameTime->m_iEndIndex = -1;	// open it back up
						//RageLog( "Re-opening the HoldStep at %d (%f)", iLastStepIndex, fLastStepBeat );
					}
					else
					{
						hs.m_TapStep = STEP_PAD2_DOWN;
						m_HoldSteps[m_iNumHoldSteps++] = hs;
						//RageLog( "Found a hold step at %d (%f).  The Step is now %d and the HoldStep is %d (%c)", iLastStepIndex, fLastStepBeat, lastStep, hs.m_TapStep, holdChar );
					}
				}
				break;
			}
			case '<':
				// hold segment
				// Handle this by just inserting one step at the end of the hold note.  
				// Then our logic below will find it an complete the hold step.

				char charStep;
				charStep = sStepData[i++];
				fCurrentBeat += 1.0/192 * BEATS_PER_MEASURE;

				while( c != '>' )	// keep reading in characters until we reach the end of the hold segment
				{
					c = sStepData[i++];
					fCurrentBeat += 1.0/192 * BEATS_PER_MEASURE;
				}

				c = charStep;	// put the step character back into 'c', and let the handler below do its magic
				// fall through!

				// hack!  undo the beat increment below
				fCurrentBeat -= fCurrentIncrementer;

			default:	// this is a note character

				TapStep step = DWIcharToStepRightSide( c );

				int iStepIndex = BeatToStepIndex( (float)fCurrentBeat );
				m_TapSteps[iStepIndex] |= step;

				//RageLog( "Setting m_TapSteps[%d] (%f) = %d  (DWI: %c)", iStepIndex, fCurrentBeat, step, c );

				// check to see if this ends one of the open HoldSteps
				for( int h=0; h<m_iNumHoldSteps; h++ )
				{
					HoldStep &hs = m_HoldSteps[h];
					if( hs.m_iEndIndex != -1 )	// this HoldStep is already finished
						continue;				// continue on to the next HoldStep

					if( step & hs.m_TapStep )	// if this step covers this hold step
					{
						hs.m_iEndIndex = iStepIndex;	// end the HoldStep here

						//RageLog( "HoldStep: step=%d, start=%d, end=%d", hs.m_TapStep, hs.m_iStartIndex, hs.m_iEndIndex );

						//RageLog( "m_TapSteps[%d]: subtracting %d from %d to make %d", iStepIndex, hs.m_TapStep, m_TapSteps[iStepIndex], m_TapSteps[iStepIndex] & ~hs.m_TapStep );
						m_TapSteps[iStepIndex] = m_TapSteps[iStepIndex] & ~hs.m_TapStep;	// subtract out the hold step
					}
				}
				
				fCurrentBeat += fCurrentIncrementer;
				break;
			}
			
		}


	}

	return true;
}


float Steps::GetLastBeat()			
{ 
	int iOldestIndexFound = 0;
	for( int i=MAX_TAP_STEP_ELEMENTS-1; i>=0; i-- )		// iterate back to front
	{
		if( m_TapSteps[i] != 0 )
			return (float)(iOldestIndexFound = i);
	}

	for( i=0; i<m_iNumHoldSteps; i++ )
	{
		if( m_HoldSteps[i].m_iEndIndex > iOldestIndexFound )
			iOldestIndexFound = m_HoldSteps[i].m_iEndIndex;
	}

	return StepIndexToBeat( (float)iOldestIndexFound );
}

int Steps::GetNumTapSteps()			
{ 
	int iNumSteps = 0;

	for( int i=0; i<MAX_TAP_STEP_ELEMENTS; i++ )
	{
		if( m_TapSteps[i] != 0 )
			iNumSteps++;
	}
	
	iNumSteps += m_iNumHoldSteps;

	return iNumSteps;
}

int Steps::GetNumHoldSteps()			
{ 
	return m_iNumHoldSteps;
}


void Steps::CropToLeftSide()
{
	TapStep pad2steps = (STEP_PAD2_LEFT | STEP_PAD2_UPLEFT | STEP_PAD2_DOWN | STEP_PAD2_UP | STEP_PAD2_UPRIGHT | STEP_PAD2_RIGHT );

	for( int i=0; i<MAX_TAP_STEP_ELEMENTS; i++ ) // foreach TapStep
	{
		// mask off the pad2 steps
		m_TapSteps[i] &= ~pad2steps;
	}

	for( i=0; i<m_iNumHoldSteps; i++ ) // foreach HoldStep
	{
		// mask off the pad2 steps
		m_HoldSteps[i].m_TapStep &= ~pad2steps;

		// If this deleted the HoldStep, remove it from the list.
		if( m_HoldSteps[i].m_TapStep == 0 )
		{
			// shift up everything below
			for( int j=i+1; j<m_iNumHoldSteps; j++ )
			{
				m_HoldSteps[j-1] = m_HoldSteps[j];
			}

			m_iNumHoldSteps--;
		}
	}

}


TapStep Pad2ToPad1( TapStep pad1step )
{
	return TapStep(pad1step >> 6);
}

void Steps::CropToRightSide()
{
	for( int i=0; i<MAX_TAP_STEP_ELEMENTS; i++ ) // foreach TapStep
	{
		// keep only the pad2 side, and shift it to the pad1 side
		m_TapSteps[i] = Pad2ToPad1( m_TapSteps[i] );
	}

	for( i=0; i<m_iNumHoldSteps; i++ ) // foreach HoldStep
	{
		// keep only the pad2 side, and shift it to the pad1 side
		m_HoldSteps[i].m_TapStep = Pad2ToPad1( m_HoldSteps[i].m_TapStep );

		// If this deleted the HoldStep, remove it from the list.
		if( m_HoldSteps[i].m_TapStep == 0 )
		{
			// shift up everything below
			for( int j=i+1; j<m_iNumHoldSteps; j++ )
			{
				m_HoldSteps[j-1] = m_HoldSteps[j];
			}

			m_iNumHoldSteps--;
		}
	}
}

void Steps::RemoveHoldSteps()
{
	// turn all the HoldSteps into TapSteps
	for( int i=0; i<m_iNumHoldSteps; i++ )
	{
		HoldStep &hs = m_HoldSteps[i];
		
		int iIndex = hs.m_iStartIndex;
		TapStep step = hs.m_TapStep;

		m_TapSteps[iIndex] |= step;
	}

	// Remove all HoldSteps
	m_iNumHoldSteps = 0;
}


void Steps::Turn( PlayerOptions::TurnType tt )
{
	// transform the steps
	TapStep oldStepToNewStep [12][2];	// 0 is old Step, 1 is new Step mapping
	oldStepToNewStep[0][0] = STEP_PAD1_LEFT;	oldStepToNewStep[0][1] = STEP_PAD1_LEFT;
	oldStepToNewStep[1][0] = STEP_PAD1_UPLEFT;	oldStepToNewStep[1][1] = STEP_PAD1_UPLEFT;
	oldStepToNewStep[2][0] = STEP_PAD1_DOWN;	oldStepToNewStep[2][1] = STEP_PAD1_DOWN;
	oldStepToNewStep[3][0] = STEP_PAD1_UP;		oldStepToNewStep[3][1] = STEP_PAD1_UP;
	oldStepToNewStep[4][0] = STEP_PAD1_UPRIGHT;	oldStepToNewStep[4][1] = STEP_PAD1_UPRIGHT;
	oldStepToNewStep[5][0] = STEP_PAD1_RIGHT;	oldStepToNewStep[5][1] = STEP_PAD1_RIGHT;
	oldStepToNewStep[6][0] = STEP_PAD2_LEFT;	oldStepToNewStep[6][1] = STEP_PAD2_LEFT;
	oldStepToNewStep[7][0] = STEP_PAD2_UPLEFT;	oldStepToNewStep[7][1] = STEP_PAD2_UPLEFT;
	oldStepToNewStep[8][0] = STEP_PAD2_DOWN;	oldStepToNewStep[8][1] = STEP_PAD2_DOWN;
	oldStepToNewStep[9][0] = STEP_PAD2_UP;		oldStepToNewStep[9][1] = STEP_PAD2_UP;
	oldStepToNewStep[10][0]= STEP_PAD2_UPRIGHT;	oldStepToNewStep[10][1]= STEP_PAD2_UPRIGHT;
	oldStepToNewStep[11][0]= STEP_PAD2_RIGHT;	oldStepToNewStep[11][1]= STEP_PAD2_RIGHT;


	switch( tt )
	{
	case PlayerOptions::TURN_NONE:
		break;
	case PlayerOptions::TURN_MIRROR:
		oldStepToNewStep[0][1] = STEP_PAD1_RIGHT;
		oldStepToNewStep[2][1] = STEP_PAD1_UP;
		oldStepToNewStep[3][1] = STEP_PAD1_DOWN;
		oldStepToNewStep[5][1] = STEP_PAD1_LEFT;
		oldStepToNewStep[6][1] = STEP_PAD2_RIGHT;
		oldStepToNewStep[8][1] = STEP_PAD2_UP;
		oldStepToNewStep[9][1] = STEP_PAD2_DOWN;
		oldStepToNewStep[11][1]= STEP_PAD2_LEFT;
		break;
	case PlayerOptions::TURN_LEFT:
		oldStepToNewStep[0][1] = STEP_PAD1_DOWN;
		oldStepToNewStep[2][1] = STEP_PAD1_RIGHT;
		oldStepToNewStep[3][1] = STEP_PAD1_LEFT;
		oldStepToNewStep[5][1] = STEP_PAD1_UP;
		oldStepToNewStep[6][1] = STEP_PAD2_DOWN;
		oldStepToNewStep[8][1] = STEP_PAD2_RIGHT;
		oldStepToNewStep[9][1] = STEP_PAD2_LEFT;
		oldStepToNewStep[11][1]= STEP_PAD2_UP;
		break;
	case PlayerOptions::TURN_RIGHT:
		oldStepToNewStep[0][1] = STEP_PAD1_UP;
		oldStepToNewStep[2][1] = STEP_PAD1_LEFT;
		oldStepToNewStep[3][1] = STEP_PAD1_RIGHT;
		oldStepToNewStep[5][1] = STEP_PAD1_DOWN;
		oldStepToNewStep[6][1] = STEP_PAD2_UP;
		oldStepToNewStep[8][1] = STEP_PAD2_LEFT;
		oldStepToNewStep[9][1] = STEP_PAD2_RIGHT;
		oldStepToNewStep[11][1]= STEP_PAD2_DOWN;
		break;
	case PlayerOptions::TURN_SHUFFLE:
		bool bAlreadyMapped[12];
		for( int i=0; i<12; i++ )
			bAlreadyMapped[i] = false;
		
		// hack: don't shuffle the Up+... Steps
		bAlreadyMapped[1] = true;
		bAlreadyMapped[4] = true;
		bAlreadyMapped[7] = true;
		bAlreadyMapped[10] = true;

		// shuffle left side
		for( i=0; i<6; i++ )
		{
			if( i == 1  ||  i == 4 )
				continue;

			int iMapsTo;
			do {
				iMapsTo = rand()%6;
			} while( bAlreadyMapped[iMapsTo] );
			bAlreadyMapped[iMapsTo] = true;

			oldStepToNewStep[i][1] = oldStepToNewStep[iMapsTo][0];
		}
		// shuffle right side
		for( i=6; i<12; i++ )
		{
			if( i == 7  ||  i == 10 )
				continue;

			int iMapsTo;
			do {
				iMapsTo = rand()%6+6;
			} while( bAlreadyMapped[iMapsTo] );
			bAlreadyMapped[iMapsTo] = true;

			oldStepToNewStep[i][1] = oldStepToNewStep[iMapsTo][0];
		}
		break;
	}

	// transform m_TapSteps 
	for( int i=0; i<MAX_TAP_STEP_ELEMENTS; i++ ) 
	{
		TapStep oldStep = m_TapSteps[i];
		TapStep &newStep = m_TapSteps[i];
		newStep = STEP_NONE;

		for( int j=0; j<12; j++ )
		{
			if( oldStep & oldStepToNewStep[j][0] )
				newStep |= oldStepToNewStep[j][1];
		}
	}

	// transform m_HoldSteps 
	for( i=0; i<m_iNumHoldSteps; i++ ) 
	{
		TapStep oldStep = m_HoldSteps[i].m_TapStep;
		TapStep &newStep = m_HoldSteps[i].m_TapStep;
		newStep = STEP_NONE;

		for( int j=0; j<12; j++ )
		{
			if( oldStep & oldStepToNewStep[j][0] )
				newStep |= oldStepToNewStep[j][1];
		}
	}
}

void Steps::MakeLittle()
{
	// filter out all non-quarter notes
	for( int i=0; i<MAX_TAP_STEP_ELEMENTS; i++ ) 
	{
		if( i%ELEMENTS_PER_BEAT != 0 )
		{
			m_TapSteps[i] = STEP_NONE;
		}
	}

	// leave HoldSteps unchanged
}



int CompareStepsPointersByFeet(const void *arg1, const void *arg2)
{
	// arg1 and arg2 are of type Step**
	Steps* pSteps1 = *(Steps**)arg1;
	Steps* pSteps2 = *(Steps**)arg2;

	int iScore1 = pSteps1->m_iNumFeet;
	int iScore2 = pSteps2->m_iNumFeet;

	if( iScore1 == iScore2 )
		return 0;
	else if( iScore1 < iScore2 )
		return -1;
	else
		return 1;
}

int CompareStepsPointersByDifficultyClass(const void *arg1, const void *arg2)
{
	// arg1 and arg2 are of type Step**
	Steps* pSteps1 = *(Steps**)arg1;
	Steps* pSteps2 = *(Steps**)arg2;

	int iDiffClass1 = pSteps1->m_DifficultyClass;
	int iDiffClass2 = pSteps2->m_DifficultyClass;

	if( iDiffClass1 == iDiffClass2 )
		return CompareStepsPointersByFeet( arg1, arg2 );
	else if( iDiffClass1 < iDiffClass2 )
		return -1;
	else
		return 1;
}

void SortStepsArrayByDifficultyClass( CArray<Steps*,Steps*> &arraySteps )
{
	qsort( arraySteps.GetData(), arraySteps.GetSize(), sizeof(Steps*), CompareStepsPointersByFeet );
}
