/*
-----------------------------------------------------------------------------
 File: Steps.h

 Desc: Hold steps data and metadata.  Does not hold dancing-time information,
	   like keeping track of which steps have been stepped on (PlayerSteps does that).

 Copyright (c) 2001 Chris Danford.  All rights reserved.
-----------------------------------------------------------------------------
*/

#ifndef _STEPS_H_
#define _STEPS_H_

#include "GameTypes.h"
#include "Grade.h"


typedef int TapStep;	// set individual bits of this to represent different notes
// step encodings
const TapStep	STEP_NONE			=	0;
const TapStep	STEP_PAD1_LEFT		=	1<<0;
const TapStep	STEP_PAD1_UPLEFT	=	1<<1;
const TapStep	STEP_PAD1_DOWN		=	1<<2;
const TapStep	STEP_PAD1_UP		=	1<<3;
const TapStep	STEP_PAD1_UPRIGHT	=	1<<4;
const TapStep	STEP_PAD1_RIGHT		=	1<<5;
const TapStep	STEP_PAD2_LEFT		=	1<<6;
const TapStep	STEP_PAD2_UPLEFT	=	1<<7;
const TapStep	STEP_PAD2_DOWN		=	1<<8;
const TapStep	STEP_PAD2_UP		=	1<<9;
const TapStep	STEP_PAD2_UPRIGHT	=	1<<10;
const TapStep	STEP_PAD2_RIGHT		=	1<<11;


enum TapStepScore { 
	TSS_NONE, 
	TSS_PERFECT, 
	TSS_GREAT,
	TSS_GOOD,
	TSS_BOO,
	TSS_MISS
};

enum TapStepTiming { 
	TST_NONE, 
	TST_EARLY, 
	TST_LATE 
};


struct HoldStep
{
	TapStep m_TapStep;	// only one bit here should be set!
	int m_iStartIndex;
	int m_iEndIndex;
};

enum HoldStepResult 
{ 
	HSR_NONE,	// this HoldStep has not been scored yet
	HSR_OK,		// the HoldStep has passed and was successfully held all the way through
	HSR_NG		// the HoldStep has passed and they missed it
};

struct HoldStepScore
{
	TapStepScore	m_TapStepScore;		// The scoring of the tap step that begins the hold.  
										// This is judeged separately from the actual hold.
	HoldStepResult	m_Result;
	float			m_fLife;	// 1.0 means this HoldStep has full life.
								// 0.0 means this HoldStep is dead
								// When this value hits 0.0 for the first time, 
								// m_HoldScore becomes HSS_NG.
								// If the life is > 0.0 when the HoldStep ends, then
								// m_HoldScore becomes HSS_OK.
	HoldStepScore() 
	{ 
		m_TapStepScore = TSS_NONE; 
		m_Result = HSR_NONE; 
		m_fLife = 1.0f; 
	};
};





const int MAX_MEASURES		= 200;	// this should be long enough to hold steps "Long Ver" songs
const int BEATS_PER_MEASURE = 4;
const int MAX_BEATS			= MAX_MEASURES * BEATS_PER_MEASURE;

const int ELEMENTS_PER_BEAT	= 12;	// It is important that this number is evenly divisible by 2, 3, 4, and 8
const int ELEMENTS_PER_MEASURE = ELEMENTS_PER_BEAT * BEATS_PER_MEASURE;

const int MAX_TAP_STEP_ELEMENTS = MAX_BEATS*ELEMENTS_PER_BEAT;

const int MAX_HOLD_STEP_ELEMENTS = 200;


inline bool IsNoteOfType( int iStepIndex, NoteType t )
{ 
	switch( t )
	{
	case NOTE_4TH:	return iStepIndex % (ELEMENTS_PER_MEASURE/4) == 0;
	case NOTE_8TH:	return iStepIndex % (ELEMENTS_PER_MEASURE/8) == 0;
	case NOTE_12TH:	return iStepIndex % (ELEMENTS_PER_MEASURE/12) == 0;
	case NOTE_16TH:	return iStepIndex % (ELEMENTS_PER_MEASURE/16) == 0;
	default:		ASSERT( false );	return false;
	}
};
inline int   BeatToStepIndex( float fBeatNum )			{ return int( fBeatNum * ELEMENTS_PER_BEAT + 0.5f); };	// round
inline int   BeatToStepIndexNotRounded( float fBeatNum ){ return (int)( fBeatNum * ELEMENTS_PER_BEAT ); };
inline float StepIndexToBeat( float fStepIndex )		{ return fStepIndex / (float)ELEMENTS_PER_BEAT;	 };
inline float StepIndexToBeat( int iStepIndex )			{ return StepIndexToBeat( (float)iStepIndex );	 };



struct Steps
{
public:
	Steps();

	// Metadata
	CString		m_sStepsFilePath;	// this is "" for file formats that don't have separate files for Steps.
	GameMode	m_GameMode;			// the GameMode that this song can be played in
	CString		m_sDescription;		// This text is displayed next to thte number of feet when a song is selected
	CString		m_sMapper;			// name of the person who created these steps
	enum DifficultyClass 
	{ 
		CLASS_EASY,		// corresponds to Basic
		CLASS_MEDIUM,	// corresponds to Trick, Another, Standard
		CLASS_HARD,		// corresponds to Maniac, SSR, Heavy
		CLASS_OTHER		// everything else
	};
	DifficultyClass m_DifficultyClass;
	int			m_iNumFeet;		// difficult of this song in "feet"

	// Data
	TapStep		m_TapSteps[MAX_TAP_STEP_ELEMENTS];
	HoldStep	m_HoldSteps[MAX_HOLD_STEP_ELEMENTS];
	int			m_iNumHoldSteps;

	// Loading
	bool LoadStepsFromBMSFile( CString sPath );
	bool LoadFromDWIValueTokens( CStringArray &arrayValueTokens );

	// Statistics
	Grade m_TopGrade;
	int m_iTopScore;
	int m_iMaxCombo;
	int m_iNumTimesPlayed;

	float GetLastBeat();	// return the beat number of the last beat in the song
	int GetNumTapSteps();
	int GetNumHoldSteps();

	// Transformations
	void CropToLeftSide();
	void CropToRightSide();
	void RemoveHoldSteps();
	void Turn( PlayerOptions::TurnType tt );
	void MakeLittle();
};



void SortStepsArrayByDifficultyClass( CArray<Steps*,Steps*> &arraySteps );


#endif