#include "stdafx.h"
/*
-----------------------------------------------------------------------------
 File: StepsSelector.h

 Desc: A graphic displayed in the StepsSelector during Dancing.

 Copyright (c) 2001 Chris Danford.  All rights reserved.
-----------------------------------------------------------------------------
*/

#include "StepsSelector.h"
#include "RageUtil.h"
#include "ScreenDimensions.h"
#include "ThemeManager.h"
#include "WindowManager.h"

const float STEPS_LINE_GAP	=	60;

StepsSelector::StepsSelector()
{
	//RageLog( "StepsSelector::StepsSelector()" );

	
	for( int i=0; i<MAX_NUM_STEPS_LINES; i++ )
	{
		m_FootMeter[i].SetXY( 0, STEPS_LINE_GAP*i );
		m_FootMeter[i].SetZoom( 1 );
		//this->AddActor( &m_FootMeter[i] );

		for( int j=0; j<NUM_PLAYERS; j++ )
		{
			m_DifficultyIcon[i][j].SetXY( (j==0 ? -250.0f : 250.0f), STEPS_LINE_GAP*i );
			m_DifficultyIcon[i][j].SetZoom( 2 );
			//this->AddActor( &m_DifficultyIcon[i][j] );
		}
	}

	m_soundChangeSteps.Load( THEME->GetPathTo(SOUND_SWITCH_STEPS) );


	for( int p=0; p<NUM_PLAYERS; p++ )
	{
		m_bEnabled[p] = false;
		m_iSelection[p] = 0;
		MakeSelectionValid( PlayerNumber(p) );
		AfterChange( (PlayerNumber)p );
		Disable( (PlayerNumber)p );
	}
}

void StepsSelector::Update( float fDeltaTime )
{
	ActorFrame::Update( fDeltaTime );


	for( int i=0; i<m_pSteps.GetSize(); i++ )
	{
		m_FootMeter[i].Update( fDeltaTime );

		for( int j=0; j<NUM_PLAYERS; j++ )
		{
			m_DifficultyIcon[i][j].Update( fDeltaTime );
		}
	}
}

void StepsSelector::RenderPrimitives()
{
	ActorFrame::RenderPrimitives();


	for( int i=0; i<m_pSteps.GetSize(); i++ )
	{
		for( int p=0; p<NUM_PLAYERS; p++ )
		{
			if( m_bEnabled[p] )
				m_DifficultyIcon[i][p].Draw();
		}

		m_FootMeter[i].Draw();
	}
}


void StepsSelector::LoadFromSong( Song* pSong, GameMode mode )
{
	if( pSong == NULL )
	{
		m_pSteps.RemoveAll();
	}
	else
	{
		for( int p=0; p<NUM_PLAYERS; p++ )
		{
			BeforeChange( (PlayerNumber)p );
		}

		m_pSteps.RemoveAll();
		
		// add some padding if all the slots aren't full
		pSong->GetStepsThatMatchGameMode( mode, m_pSteps );
		SortStepsArrayByDifficultyClass( m_pSteps );
		if( m_pSteps[0]->m_DifficultyClass == Steps::CLASS_MEDIUM )
		{
			m_pSteps.InsertAt( 0, (Steps*)NULL );
		}
		else if( m_pSteps[0]->m_DifficultyClass == Steps::CLASS_HARD )
		{
			m_pSteps.InsertAt( 0, (Steps*)NULL );
			m_pSteps.InsertAt( 0, (Steps*)NULL );
		}


		// load new icons and set the foot meters
		for( int i=0; i<MAX_NUM_STEPS_LINES && i<m_pSteps.GetSize(); i++ )
		{
			m_FootMeter[i].SetFromSteps( m_pSteps[i] );

			for( int p=0; p<NUM_PLAYERS; p++ )
			{
				m_DifficultyIcon[i][p].SetFromSteps( m_pSteps[i] ); 
			}
		}


		for( p=0; p<NUM_PLAYERS; p++ )
		{
			MakeSelectionValid( (PlayerNumber)p );
			AfterChange( (PlayerNumber)p );
		}
	}
	WM->SendMessageToTopWindow( SM_StepsChanged, 0 );
}

void StepsSelector::Up( PlayerNumber p )
{
	RageLog( "StepsSelector::Up( %d )", p );

	if( !m_bEnabled[p] )
		return;
	if( m_pSteps.GetSize() == 0 )
		return;
	if( m_iSelection[p] == 0 )
		return;

	for( int i=m_iSelection[p]-1; i>=0; i-- )
	{
		if( m_pSteps[i] != NULL )
			break;	// we found one!
	}
	if( i == -1 )
		return;
		

	ChangeSelection( p, i );
}

void StepsSelector::Down( PlayerNumber p )
{
	RageLog( "StepsSelector::Down( %d )", p );

	if( !m_bEnabled[p] )
		return;
	if( m_pSteps.GetSize() == 0 )
		return;
	if( m_iSelection[p] == m_pSteps.GetSize()-1 )
		return;

	for( int i=m_iSelection[p]+1; i<m_pSteps.GetSize(); i++ )
	{
		if( m_pSteps[i] != NULL )
			break;	// we found one!
	}
	if( i == m_pSteps.GetSize() )
		return;


	ChangeSelection( p, i );
}

void StepsSelector::ChangeSelection( PlayerNumber p, int iNewSelection )
{
	BeforeChange( p );

	m_iSelection[p] = iNewSelection;

	AfterChange( p );
	WM->SendMessageToTopWindow( SM_StepsChanged, 0 );
	m_soundChangeSteps.PlayRandom();
}

void StepsSelector::BeforeChange( PlayerNumber p )
{
	int iSelectedLine = m_iSelection[p];
	m_DifficultyIcon[iSelectedLine][p].SetEffectNone();
}

void StepsSelector::AfterChange( PlayerNumber p )
{
	D3DXVECTOR3 vectBouncDir;
	
	int iSelectedLine = m_iSelection[p];
	m_DifficultyIcon[iSelectedLine][p].SetEffectNone();

	switch( p )
	{
	case PLAYER_1:	vectBouncDir = D3DXVECTOR3(-50,0,0);	break;
	case PLAYER_2:	vectBouncDir = D3DXVECTOR3(50,0,0);		break;
	default:		ASSERT( false );							break;
	}

	m_DifficultyIcon[iSelectedLine][p].SetEffectBouncing( vectBouncDir, 0.7f );
}



