#include "stdafx.h"
/*
-----------------------------------------------------------------------------
 Class: WindowConfigurePads

 Desc: Where the player maps device input to pad input.

 Copyright (c) 2001 Chris Danford.  All rights reserved.
-----------------------------------------------------------------------------
*/

#include "WindowConfigurePads.h"
#include <assert.h>
#include "PrefsManager.h"
#include "WindowManager.h"
#include "WindowTitleMenu.h"
#include "ScreenDimensions.h"
#include "ThemeManager.h"


const float PAD_X[NUM_PADS]	= { SCREEN_WIDTH*1/4, SCREEN_WIDTH*3/4 };

const float TITLE_Y			=	30;
const float HELP_X			=	CENTER_X;
const float HELP_Y			=	SCREEN_HEIGHT-20;
const float LINE_START_Y	=	60;
const float LINE_GAP_Y		=	50;



WindowConfigurePads::WindowConfigurePads()
{
	RageLog( "WindowConfigurePads::WindowConfigurePads()" );
	
	m_iCurPadNum = (PadNumber)0;
	m_iCurButton = (PadButton)0;
	m_iCurSlot = 0;

	m_bWaitingForPress = false;

	InitText();
	InitGraphicalNames();
	Refresh();

	m_fFakeSongBeat = 0;

	m_fErrorDisplayCountdown = 0;
}



WindowConfigurePads::~WindowConfigurePads()
{
	RageLog( "WindowConfigurePads::~WindowConfigurePads()" );
}


void WindowConfigurePads::Update( float fDeltaTime )
{
	m_textTitle.Update( fDeltaTime );
	m_textHelp.Update( fDeltaTime );
	m_textError.Update( fDeltaTime );

	m_fErrorDisplayCountdown -= fDeltaTime;
	if( m_fErrorDisplayCountdown < 0 )
		m_fErrorDisplayCountdown = 0;

	m_fFakeSongBeat += 50.0f/60 * fDeltaTime;	// assume 50 BPM

	for( int b=0; b<NUM_PAD_BUTTONS; b++ )
	{
		for( int p=0; p<NUM_PADS; p++ )
			for( int n=0; n<NUM_SLOTS_SHOWN; n++ ) 
				m_textMappedTo[p][b][n].Update( fDeltaTime );

		m_textName[b].Update( fDeltaTime );
		m_GrayArrowName[b].SetBeat( m_fFakeSongBeat );
	}
}


void WindowConfigurePads::RenderPrimitives()
{
	m_textTitle.Draw();
	m_textHelp.Draw();

	for( int b=0; b<NUM_PAD_BUTTONS; b++ )
	{
		for( int p=0; p<NUM_PADS; p++ ) 
			for( int n=0; n<NUM_SLOTS_SHOWN; n++ )
				m_textMappedTo[p][b][n].Draw();
		
		if( b<6 ) {
			m_GrayArrowName[b].Draw();
		} else {
			m_textName[b].Draw();
		}
	}

	if( m_fErrorDisplayCountdown > 0 )
		m_textError.Draw();
}


void WindowConfigurePads::Input( DeviceInput& DeviceI, PadInput &PadI, PlayerInput &PlayerI )
{
	RageLog( "WindowConfigurePads::Input():  device: %d, button: %d", DeviceI.device, DeviceI.button );

	if( m_bWaitingForPress )	// we're going to map an input
	{	
		// ignore joystick D-Pad presses if the user has set their pref.
		if( PREFS->m_GameOptions.m_bIgnoreJoyAxes  &&
			DEVICE_JOYSTICK1 <= DeviceI.device  &&  DeviceI.device <= DEVICE_JOYSTICK4  &&
			( DeviceI.button == JOY_LEFT || DeviceI.button == JOY_RIGHT || DeviceI.button == JOY_UP || DeviceI.button == JOY_DOWN ) )
		{
			//m_textError.SetText( "Game option is set to ignore the Joystick D-Pad." );
			//m_fErrorDisplayCountdown = 5;	// show the error message
			return;	// ignore this press
		}
		else if( PREFS->IsAHardCodedDeviceInput( DeviceI ) )
		{
			m_textError.SetText( "Sorry.  This special key cannot be re-mapped." );
			m_fErrorDisplayCountdown = 5;	// show the error message
		}
		else
		{
			PadInput curPadI( (PadNumber)m_iCurPadNum, (PadButton)m_iCurButton );

			PREFS->SetInputMap( DeviceI, curPadI, m_iCurSlot );
			
			PREFS->SetHardCodedButtons();

			PREFS->SavePrefsToDisk();	// commit to disk so we don't lose the changes!
		}

		m_bWaitingForPress = false;
	}
	else if( DeviceI.device == DEVICE_KEYBOARD  &&  DeviceI.button == DIK_SPACE )
	{
		PadInput curPadI( (PadNumber)m_iCurPadNum, (PadButton)m_iCurButton );

		PREFS->ClearFromInputMap( curPadI, m_iCurSlot );

		PREFS->SetHardCodedButtons();

		PREFS->SavePrefsToDisk();	// commit to disk so we don't lose the changes!
	}
	else
	{
		Window::Input( DeviceI, PadI, PlayerI );	// default handler
	}

	Refresh();
}

void WindowConfigurePads::HandleWindowMessage( WindowMessage SM )
{
}

void WindowConfigurePads::InitText()
{

	m_textTitle.Load( THEME->GetPathTo(FONT_NORMAL) );
	m_textTitle.SetText( "Configure Pads" );
	m_textTitle.SetXY( CENTER_X, TITLE_Y );
	m_textTitle.SetZoom( 1.2f );

	m_textHelp.Load( THEME->GetPathTo(FONT_NORMAL) );
	m_textHelp.SetText( "Use arrow keys to navigate, ENTER to assign, and SPACE to clear." );
	m_textHelp.SetXY( HELP_X, HELP_Y );
	m_textHelp.SetZoom( 0.6f );

	m_textError.Load( THEME->GetPathTo(FONT_NORMAL) );
	m_textError.SetText( "" );
	m_textError.SetXY( CENTER_X, CENTER_Y );
	m_textError.SetDiffuseColor( D3DXCOLOR(0,1,0,1) );
	m_textError.SetZoom( 0.8f );

	for( int b=0; b<NUM_PAD_BUTTONS; b++ ) {

		m_textName[b].Load( THEME->GetPathTo(FONT_NORMAL) );
		m_textName[b].SetXY( CENTER_X, LINE_START_Y + b*LINE_GAP_Y + 8);
		m_textName[b].SetText( g_sPIButtonDescription[b] );
		m_textName[b].SetZoom( 0.8f );
		
		for( int p=0; p<NUM_PADS; p++ ) {			
			for( int n=0; n<NUM_SLOTS_SHOWN; n++ ) {

				m_textMappedTo[p][b][n].Load( THEME->GetPathTo(FONT_NORMAL) );
				m_textMappedTo[p][b][n].SetXY( PAD_X[p], LINE_START_Y + b*LINE_GAP_Y + 8 + n*15 );
				m_textMappedTo[p][b][n].SetText( "(none)" );
				m_textMappedTo[p][b][n].SetZoom( 0.5f );
			}
		}
	}	
}	

void WindowConfigurePads::InitGraphicalNames()
{
	float fRotation[NUM_PAD_BUTTONS] = {
		0,
		D3DX_PI/4.0f,
		-D3DX_PI/2.0f,
		D3DX_PI/2.0f,
		D3DX_PI*3.0f/4.0f,
		D3DX_PI,
		0,	// back
		0	// next
	};

	for( int i=0; i<NUM_PAD_BUTTONS; i++ )
	{
		m_GrayArrowName[i].SetXY( CENTER_X, LINE_START_Y + i*LINE_GAP_Y + 8);
		m_GrayArrowName[i].SetRotation( fRotation[i] );
		m_GrayArrowName[i].SetZoom( 0.5f );
	}
}

void WindowConfigurePads::Refresh()
{
	for( int p=0; p<NUM_PADS; p++ ) 
	{			
		for( int b=0; b<NUM_PAD_BUTTONS; b++ ) 
		{
			for( int s=0; s<NUM_SLOTS_SHOWN; s++ ) 
			{
				PadInput cur_pi( (PadNumber)p, (PadButton)b );
				DeviceInput di;
				if( PREFS->PadToDevice( cur_pi, s, di ) )
					m_textMappedTo[p][b][s].SetText( di.GetDescription() );
				else
					m_textMappedTo[p][b][s].SetText( "-----------" );
				
				// highlight the currently selected pad button
				if( p == m_iCurPadNum  &&  b == m_iCurButton  &&  s == m_iCurSlot ) 
				{
					if( m_bWaitingForPress )
						m_textMappedTo[p][b][s].SetDiffuseColor( D3DXCOLOR(1,0.5,0.5,1) );	// red
					else
						m_textMappedTo[p][b][s].SetDiffuseColor( D3DXCOLOR(1,1,1,1) );	// white
				} 
				else 
				{
					m_textMappedTo[p][b][s].SetDiffuseColor( D3DXCOLOR(0.5,0.5,0.5,1) );	// gray
				}
			}
		}
	}
}

void WindowConfigurePads::Left()
{
	m_iCurPadNum = (PadNumber)0;
}

void WindowConfigurePads::Right()
{
	m_iCurPadNum = (PadNumber)1;
}

void WindowConfigurePads::Up()
{
	if( m_iCurSlot == 0 )
	{
		m_iCurButton--;
		if( m_iCurButton < 0 )
			m_iCurButton = 0;

		m_iCurSlot = 1;
	}
	else if( m_iCurSlot == 1 )
	{
		m_iCurSlot = 0;
	}
}

void WindowConfigurePads::Down()
{
	if( m_iCurSlot == 0 )
	{
		m_iCurSlot = 1;
	}
	else if( m_iCurSlot == 1 )
	{
		m_iCurButton++;
		if( m_iCurButton > NUM_PAD_BUTTONS-1 )
			m_iCurButton = NUM_PAD_BUTTONS-1;

		m_iCurSlot = 0;
	}
}

void WindowConfigurePads::Back()
{
	WM->SetNewWindow( new WindowTitleMenu );
}

void WindowConfigurePads::Next()
{
	m_bWaitingForPress = true;
}

