#include "stdafx.h"
/*
-----------------------------------------------------------------------------
 File: WindowManager.h

 Desc: Manages the game windows.

 Copyright (c) 2001 Chris Danford.  All rights reserved.
-----------------------------------------------------------------------------
*/

#include "WindowManager.h"
#include "IniFile.h"
#include "ScreenDimensions.h"
#include "ThemeManager.h"


WindowManager*	WM = NULL;	// global and accessable from anywhere in our program



WindowManager::WindowManager()
{
	m_textFPS.Load( THEME->GetPathTo(FONT_NORMAL) );
	m_textFPS.SetXY( SCREEN_WIDTH-35, 15 );
	m_textFPS.SetZ( -2 );
	m_textFPS.SetZoom( 0.5f );

	for( int i=0; i<MAX_NUM_SYSTEM_MESSAGE_LINES; i++ )
	{
		m_textSystemMessage[i].Load( THEME->GetPathTo(FONT_NORMAL) );
		m_textSystemMessage[i].SetHorizAlign( Actor::align_left );
		m_textSystemMessage[i].SetVertAlign( Actor::align_bottom );
		m_textSystemMessage[i].SetXY( 5.0f, 10.0f+20*i );
		m_textSystemMessage[i].SetZ( -2 );
		m_textSystemMessage[i].SetZoom( 0.5f );
	}
}


WindowManager::~WindowManager()
{
	RageLog( "WindowManager::~WindowManager()" );

	// delete current states
	for( int i=0; i<m_WindowStack.GetSize(); i++ )
		SAFE_DELETE( m_WindowStack[i] );
}




void WindowManager::Update( float fDeltaTime )
{
	for( int i=0; i<MAX_NUM_SYSTEM_MESSAGE_LINES; i++ )
	{
		m_textSystemMessage[i].Update( fDeltaTime );
	}
	m_textFPS.Update( fDeltaTime );

	// delete all WindowsToDelete
	for( i=0; i<m_WindowsToDelete.GetSize(); i++ ) {
		SAFE_DELETE( m_WindowsToDelete[i] );
		m_WindowsToDelete.RemoveAt(i);
	}

	// HACK!  If we deleted at least one state, then skip this update!
	if( i>0 ) return;

	// Update all windows in the stack
	for( i=0; i<m_WindowStack.GetSize(); i++ )
		m_WindowStack[i]->Update( fDeltaTime );

}


void WindowManager::Restore()
{
	// Draw all CurrentWindows (back to front)
	for( int i=0; i<m_WindowStack.GetSize(); i++ )
		m_WindowStack[i]->Restore();
}

void WindowManager::Invalidate()
{
	for( int i=0; i<m_WindowStack.GetSize(); i++ )
		m_WindowStack[i]->Invalidate();
}

void WindowManager::Draw()
{
	// Draw all CurrentWindows (back to front)
	for( int i=0; i<m_WindowStack.GetSize(); i++ )
		m_WindowStack[i]->Draw();
	
	for( i=0; i<MAX_NUM_SYSTEM_MESSAGE_LINES; i++ )
	{
		m_textSystemMessage[i].Draw();
	}

	if( PREFS  &&  PREFS->m_GameOptions.m_bShowFPS )
	{
		m_textFPS.SetText( ssprintf("%3.0f FPS", SCREEN->GetFPS()) );
		m_textFPS.Draw();
	}
}


void WindowManager::Input( DeviceInput& DeviceI, PadInput &PadI, PlayerInput &PlayerI )
{
	RageLog( "WindowManager::Input( %d-%d, %d-%d, %d-%d )", 
		DeviceI.device, DeviceI.button, PadI.pad_no, PadI.button, PlayerI.player_no, PlayerI.step );

	// pass input only to topmost state
	if( m_WindowStack.GetSize() > 0 )
		m_WindowStack[m_WindowStack.GetSize()-1]->Input( DeviceI, PadI, PlayerI );
}

void WindowManager::SetNewWindow( Window *pNewWindow )
{
	// move CurrentWindow to WindowToDelete
	m_WindowsToDelete.Copy( m_WindowStack );

	m_WindowStack.RemoveAll();
	m_WindowStack.Add( pNewWindow );
}

void WindowManager::AddWindowToTop( Window *pNewWindow )
{
	// our responsibility to tell the old state that it's losing focus
	m_WindowStack[m_WindowStack.GetSize()-1]->HandleWindowMessage( SM_LosingInputFocus );

	// add the new state onto the back of the array
	m_WindowStack.Add( pNewWindow );
}

void WindowManager::PopTopWindow()
{
	Window* pWindowToPop = m_WindowStack[m_WindowStack.GetSize()-1];	// top menu
	//pWindowToPop->HandleWindowMessage( SM_LosingInputFocus );
	m_WindowStack.RemoveAt(m_WindowStack.GetSize()-1);
	m_WindowsToDelete.Add( pWindowToPop );
	
	Window* pNewTopWindow = m_WindowStack[m_WindowStack.GetSize()-1];
	pNewTopWindow->HandleWindowMessage( SM_RegainingInputFocus );
}

void WindowManager::SendMessageToTopWindow( WindowMessage SM, float fDelay )
{
	Window* pTopWindow = m_WindowStack[m_WindowStack.GetSize()-1];
	pTopWindow->SendWindowMessage( SM, fDelay );
}


void WindowManager::SystemMessage( CString sMessage )
{
	// Look for an open spot
	for( int i=0; i<MAX_NUM_SYSTEM_MESSAGE_LINES; i++ )
	{
		if( m_textSystemMessage[i].GetDiffuseColor().a == 0  ||  m_textSystemMessage[i].GetText() == "" )
		{
			m_textSystemMessage[i].StopTweening();
			m_textSystemMessage[i].SetText( sMessage );
			m_textSystemMessage[i].SetDiffuseColor( D3DXCOLOR(1,1,1,1) );
			m_textSystemMessage[i].BeginTweeningQueued( 5 );
			m_textSystemMessage[i].BeginTweeningQueued( 1 );
			m_textSystemMessage[i].SetTweenDiffuseColor( D3DXCOLOR(1,1,1,0) );

			break;	// done!
		}

	}
}


